@@ -1,27 +1,147 @@
 # -*- coding: utf-8 -*-
-
+import logging
 from copy import deepcopy
 import pickle, time
 from shutil import copy
 import sys, os
 import warnings
 
-from joblib._multiprocessing_helpers import mp
-import matplotlib
 import powerlaw
-from scipy import optimize
-from scipy.linalg import svd
+from scipy import optimize, stats
 from sklearn.neighbors import KernelDensity
 import tqdm
 
 import matplotlib.pyplot as plt
 import numpy as np
-import pandas as pd
 import scipy as sp
 import scipy.stats as stats
 
-
 from .constants import *
+from .WW_powerlaw import *
+
+
+logger = logging.getLogger(WW_NAME)
+
+
+def has_mac_accelerate():
+    """Check if the platform is the MAC M1/M2 and numpy is using the accelerate library.
+    If so, numpy SVD will be faster than linalg SVD, and we auto-switch to that upon import
+    
+    Note: """
+    
+    import platform
+    import warnings
+    
+    use_linalg = False
+    mac_arch = platform.machine()
+    if mac_arch == 'arm64':
+    
+        with warnings.catch_warnings():
+            warnings.filterwarnings("ignore",category=DeprecationWarning)
+            import numpy.distutils.system_info as sysinfo
+            info = sysinfo.get_info('accelerate')
+            if info is not None and len(info)>0:
+                for x in info['extra_link_args']:
+                    if 'Accelerate' in x:
+                        use_linalg = True
+                        
+    return use_linalg
+    
+
+
+# PyTorch has 2 separate SVD methods, one for the vals, and one for the factorization
+# https://pytorch.org/docs/stable/generated/torch.linalg.eig.html
+# https://pytorch.org/docs/stable/generated/torch.linalg.svd.html
+# https://pytorch.org/docs/stable/generated/torch.linalg.svdvals.html
+
+# TODO: change this based on ENV var and/or auto-detect
+# see: issue #220
+if has_mac_accelerate():
+    logger.info("Using Numoy with MAC M1/2 Accelerate for SVD")
+    _eig_full_accurate = lambda W: np.linalg.eig(W)
+    _svd_full_accurate = lambda W: np.linalg.svd(W, compute_uv=True)
+    _svd_vals_accurate = lambda W: np.linalg.svd(W, compute_uv=False)
+else:
+    logger.info("Using Scipy for SVD")
+    _eig_full_accurate = lambda W: sp.linalg.eig(W)
+    _svd_full_accurate = lambda W: sp.linalg.svd(W, compute_uv=True)
+    _svd_vals_accurate = lambda W: sp.linalg.svd(W, compute_uv=False)
+    
+    
+    
+def has_cuda():
+    try:
+        import torch
+        return torch.cuda.is_available()
+    except ImportError:
+        pass
+    return False
+
+
+"""This method is over-written in tourch cuda is available"""
+torch_T_to_np = lambda T: T.to("cpu").float().numpy()
+
+# 
+try:
+    import torch
+    if torch.cuda.is_available() :
+        logger.info("Torch CUDA available, using torch_wrapper")
+
+        # float changed to half to save memory
+        torch_T_to_np = lambda T: T.to("cpu").half().numpy()
+         
+        def torch_wrapper(M, f):
+            torch.cuda.empty_cache()
+            with torch.no_grad():
+                M_cuda = torch.Tensor(M).to("cuda")
+                rvals = f(M_cuda)
+            del M_cuda
+            return rvals          
+            
+
+        def _eig_full_fast(M):
+            L, V = torch_wrapper(M, lambda M: torch.linalg.eig(M))
+            return torch_T_to_np(L), torch_T_to_np(V)
+        def _svd_full_fast(M):
+            U, S, Vh = torch_wrapper(M, lambda M: torch.linalg.svd(M))
+            return torch_T_to_np(U), torch_T_to_np(S), torch_T_to_np(Vh)
+        def _svd_vals_fast(M):
+            S = torch_wrapper(M, lambda M: torch.linalg.svdvals(M))
+            return torch_T_to_np(S)
+    else:
+        msg_svd = "SciPy"
+        if has_mac_accelerate():
+            msg_svd = "NumPy"
+
+        logger.warning(f"PyTorch is available but CUDA is not. Defaulting to {msg_svd} for SVD")
+        raise ImportError()
+    
+except ImportError:
+    # if torch / cuda are not available, default to scipy
+    _eig_full_fast = _eig_full_accurate
+    _svd_full_fast = _svd_full_accurate
+    _svd_vals_fast = _svd_vals_accurate
+    
+    torch_T_to_np = lambda T: T.to("cpu").float().numpy()
+
+    
+
+def eig_full(W, method=ACCURATE_SVD):
+    assert method.lower() in [ACCURATE_SVD, FAST_SVD], method #TODO TRUNCATED_SVD?
+    if method == ACCURATE_SVD: return _eig_full_accurate(W)
+    if method == FAST_SVD:     return _eig_full_fast(W)
+
+def svd_full(W, method=ACCURATE_SVD):
+    assert method.lower() in [ACCURATE_SVD, FAST_SVD], method #TODO TRUNCATED_SVD
+    if method == ACCURATE_SVD: return _svd_full_accurate(W)
+    if method == FAST_SVD:     return _svd_full_fast(W)
+
+def svd_vals(W, method=ACCURATE_SVD):
+    assert method.lower() in [ACCURATE_SVD, FAST_SVD], method #TODO TRUNCATED_SVD
+    if method == ACCURATE_SVD: return _svd_vals_accurate(W)
+    if method == FAST_SVD:     return _svd_vals_fast(W)
+
+
 
 
 # ## Generalized Entropy
@@ -73,21 +193,24 @@ def marchenko_pastur_pdf(x_min, x_max, Q, sigma=1.0, grid_size=0.001):
     """Return a discrete set of (x,y) values representing the Marchenko-Pastur distrbution from RMT.
     Computing on a grid of size 0.001 by default"""
     y = 1 / Q
-    x = np.arange(x_min, x_max, grid_size)
 
     b = np.power(sigma * (1 + np.sqrt(1 / Q)), 2)  # Largest eigenvalue
     a = np.power(sigma * (1 - np.sqrt(1 / Q)), 2)  # Smallest eigenvalue
-    
+
+    x_min = max(x_min, a)
+    x_max = min(x_max, b)
+    x = np.arange(x_min, x_max, grid_size)
+
     return x, (1 / (2 * np.pi * sigma * sigma * x * y)) * np.sqrt((b - x) * (x - a))
 
 
 def quarter_circle_pdf(x_min, x_max, sigma=1.0, grid_size=0.001):
     """Return a discrete set of (x,y) values representing the Quarter-Circle distribution from RMT.
     Computing on a grid of size 0.001 by default"""
-    x = np.arange(x_min, x_max, grid_size)
-    
+    x = np.arange(x_min, min(x_max, 2), grid_size)
+
     # When Q = 1, we look at the singular values instaed of eigenvalues
-    return x, (1 / (np.pi * sigma * sigma)) * np.sqrt(4 - x ** 2)
+    return x, (1 / (np.pi * sigma * sigma)) * np.sqrt((4 - x ** 2)+10**-9)
 
 
 def calc_sigma(Q, evals):
@@ -126,7 +249,7 @@ def hard_rank(W, sv):
 
 
 # uss FAST SVD method: notice we miss 1 eigenvalue here...using 
-def get_shuffled_eigenvalues(W, layer=7, num=100):
+def get_shuffled_eigenvalues(W, layer=7, num=100, method=ACCURATE_SVD):
     "get eigenvalues for this model, but shuffled, num times"
     
     N, M = W.shape[0], W.shape[1]       
@@ -140,7 +263,7 @@ def get_shuffled_eigenvalues(W, layer=7, num=100):
         np.random.shuffle(W_shuf)
         W_shuf = W_shuf.reshape([N, M])
 
-        u, sv, vh = svd(W_shuf)
+        sv = svd_vals(W_shuf, method=method)
 
         eigenvalues.extend(sv * sv)
         
@@ -190,7 +313,7 @@ def plot_density_and_fit(eigenvalues=None, model=None, layer_name="", layer_id=0
     percent_mass = 100.0 * (num_spikes) / len(to_fit)
 
     bulk_fit = np.sort(to_fit)[num_spikes:]
-    x_min, x_max = 0, np.max(bulk_fit)
+    x_min, x_max = np.min(bulk_fit), np.max(bulk_fit)
     
     if Q == 1:
         x, mp = quarter_circle_pdf(x_min, x_max, sigma)
@@ -220,12 +343,12 @@ def plot_density(to_plot, sigma, Q, method="MP", color='blue', cutoff=0.0):
 
     if method == "MP":
         to_plot = np.sort(to_plot)
-        x_min, x_max = 0, np.max(to_plot)
+        x_min, x_max = np.min(to_plot), np.max(to_plot)
         x, y = marchenko_pastur_pdf(x_min, x_max, Q, sigma)
     elif method == "QC":
         cutoff = np.sqrt(cutoff)
         to_plot = np.sort(to_plot)
-        x_min, x_max = 0, np.max(to_plot)
+        x_min, x_max = np.min(to_plot), np.max(to_plot)
         x, y = quarter_circle_pdf(x_min, x_max, sigma)
         
     plt.hist(to_plot, bins=100, alpha=0.6, color=color, density=True, label="ead")
@@ -268,11 +391,11 @@ def stable_rank(evals):
     return np.sum(evals) / np.max(evals)
 
 
-def matrix_soft_rank(W):
+def matrix_soft_rank(W, method=ACCURATE_SVD):
     """compute the matrix soft rank (or stable rank), given rectangular numpy matrix W""
     See:  https://arxiv.org/abs/1810.01075 """ 
     W = W / np.trace(W)
-    u, sv, vh = svd(W)
+    sv = svd_vals(W, method=method)
     return stable_rank(sv * sv)
 
 
@@ -354,6 +477,21 @@ def max_discrete_entropy(len_vec, num_bins=100, sample_size=100000):
 # ## Misc
 
 
+def line_entropy(data):
+    """Compute the entropy of a 1d function
+    
+    see issue #173 """
+    
+    # Discretize the data
+    hist, bin_edges = np.histogram(data, bins='auto')
+    # Compute the probability density of the function
+    prob = hist / np.sum(hist)
+    # Compute the entropy of the function
+    the_entropy = stats.entropy(prob)
+    return the_entropy
+
+
+
 def spike_min(Q):
     """minimum perturbation to W to see a spike"""
     return 1 / np.sqrt(np.sqrt(Q))
@@ -553,6 +691,7 @@ def fit_density_with_range(evals, Q, bw=0.1, sigma_range=(slice(0.1, 1.25, 0.01)
     
     return brute_output[0][0], brute_output[1]  # sigma_optimized, resid
 
+# import pandas as pd
 # def fit_mp_findspikes(evals, Q):
 #     '''Remove eigen (spikes) from largest to smallest'''
 #     evals = sorted(evals)[::-1]
@@ -614,8 +753,8 @@ def save_fig(plt, figname, layer_id, savedir):
     plt.tight_layout()
     plt.savefig(figname)
     return 
-
-#TODO:
+ 
+#TODO: WHAT IS THIS?
 def fit_xxx_powerlaw(evals, xmin=None):
     N = int(len(evals)/4)
     xmax = np.max(evals)       
@@ -626,7 +765,6 @@ def fit_xxx_powerlaw(evals, xmin=None):
         prev_fit = powerlaw.Fit(evals, xmax=xmax)
        
     fit = prev_fit 
-    print(f"fit alpha {fit.alpha:0.2f} sigma {fit.sigma:0.2f}")  
 
     prev_alpha = prev_fit.alpha
     prev_sigma = prev_fit.sigma
@@ -637,7 +775,9 @@ def fit_xxx_powerlaw(evals, xmin=None):
     return fit
     
 # check alpha is decreasing
-def fit_clipped_powerlaw(evals, xmin=None, verbose=False, max_N=DEFAULT_MAX_N, min_alpha=2.0, alpha_thresh=1.0, logger=None, plot=False):
+# DO WE NEED XMAX='force'?
+def fit_clipped_powerlaw(evals, xmin=None, xmax=None, verbose=False, max_fingers=DEFAULT_MAX_FINGERS, min_alpha=2.0, 
+                         finger_thresh=DEFAULT_FINGER_THRESH, logger=None, plot=False, pl_package=WW_POWERLAW_PACKAGE):
     """Fits a powerlaw only, not a truncated power law
        clips off the max evals until a powerlaw is found, or stops half-way into the ESD
        
@@ -647,11 +787,11 @@ def fit_clipped_powerlaw(evals, xmin=None, verbose=False, max_N=DEFAULT_MAX_N, m
        
        Parameters:
        
-         max_N:  max number of eigenvakues to clip
+         max_fingers:  max number of eigenvalues to clip
        
          min_alpha:  stops if alpha drops below min_alpha
          
-         alpha_thresh=1.0   alpha has to drop this much to stop
+         finger_thresh=1.0   alpha has to drop this much to stop
         
         NOTE: currently this only works for the fit='power law'
        
@@ -662,90 +802,101 @@ def fit_clipped_powerlaw(evals, xmin=None, verbose=False, max_N=DEFAULT_MAX_N, m
         logger = logging.getLogger("RMT_Util")
         logger.setLevel(logging.INFO)
                 
-    logger.info(f"fit_clipped_powerlaw {max_N} ")
+    logger.info(f"fit_clipped_powerlaw: max_fingers={MAX_FINGERS} xmax={xmax}")
     fit = None
     #with warnings.catch_warnings():
         #warnings.simplefilter(action='ignore', category=RuntimeWarning) 
             
     #assert(evals[-1]> evals[0]) 
-    xmax = np.max(evals)
+    # xmaxxmax = None #np.max(evals)
+
+    if xmin is not None and xmin != -1:
+        #prev_fit =  powerlaw.Fit(evals, xmin=xmin, xmax=xmax, verbose=verbose)
+        prev_fit =  pl_fit(data=evals, xmin=xmin, xmax=xmax, verbose=verbose, distribution=POWER_LAW, pl_package=pl_package)
+    else:
+        #prev_fit = powerlaw.Fit(evals, xmax=xmax)
+        prev_fit =  pl_fit(data=evals, xmax=xmax, verbose=verbose, distribution=POWER_LAW, pl_package=pl_package)
 
     
-    if xmin is not None and xmin != -1:
-        prev_fit = powerlaw.Fit(evals, xmin=xmin, xmax=xmax, verbose=verbose)
-    else:
-        prev_fit = powerlaw.Fit(evals, xmax=xmax)
-        
-    R, p = prev_fit.distribution_compare('truncated_power_law', 'power_law', normalized_ratio=True)
-    logger.info(f"fit alpha {prev_fit.alpha:0.2f} sigma {prev_fit.sigma:0.2f} TPL or PL? {R:0.4f}")     
+    # NOT USED ANYMORE
+    #R = 1.0
+    #p = 0.01
+    #R, p = prev_fit.distribution_compare('truncated_power_law', 'power_law', normalized_ratio=True)
+    #logger.info(f"fit alpha {prev_fit.alpha:0.2f} sigma {prev_fit.sigma:0.2f} TPL or PL? {R:0.4f}")     
 
     prev_alpha = prev_fit.alpha
     prev_sigma = prev_fit.sigma
-    prev_R = R
+    #prev_R = R
     first_fit = prev_fit
     best_fit = first_fit
 
     num_fingers = 0
     
-    for idx in range(2,max_N):
-        xmax = np.max(evals[-idx])
+    for idx in range(2,max_fingers):
+        if xmax is not None and xmax is not False:
+            xmax = np.max(evals[-idx])
+        
          
         if xmin is not None and xmin != -1:
-            fit = powerlaw.Fit(evals, xmin=xmin, xmax=xmax, verbose=verbose)
+            #fit = powerlaw.Fit(evals, xmin=xmin, xmax=xmax, verbose=verbose)
+            fit = pl_fit(data=evals[:-idx], xmin=xmin, xmax=xmax, verbose=verbose, distribution=POWER_LAW, pl_package=pl_package)
         else:
-            fit = powerlaw.Fit(evals, xmax=xmax, verbose=verbose)
+            #fit = powerlaw.Fit(evals, xmax=xmax, verbose=verbose)
+            fit = pl_fit(data=evals[:-idx], xmax=xmax, verbose=verbose, distribution=POWER_LAW, pl_package=pl_package)
   
         
         # this is only meaningful if the orginal fit is a TPL and the new fit is a PL
         # stop when distribution becomes power law
-        R, p = fit.distribution_compare('truncated_power_law', 'power_law', normalized_ratio=True)
-        logger.info(f"{idx} fit alpha {fit.alpha:0.4f} sigma {fit.sigma:0.4f} TPL or PL? {R:0.4f}")     
+        #R, p = fit.distribution_compare('truncated_power_law', 'power_law', normalized_ratio=True)
+        logger.info(f"fix_fingers: idx={idx} fit alpha {fit.alpha:0.4f}  sigma {fit.sigma:0.4f} xmax={xmax}")     
         #if R > 0.0:
         #    break
    
-        
-        logger.info(f"{idx} fit alpha {fit.alpha:0.2f} sigma {fit.sigma:0.2f} TPL or PL? {R:0.4f}")     
-
-        if ((fit.alpha + alpha_thresh) < first_fit.alpha) : #and fit.sigma < prev_sigma:
-            logger.info(f"stopping at {idx} {fit.alpha:.2f} << {prev_alpha:0.2f} ")  
+        if ((fit.alpha + finger_thresh) < first_fit.alpha) : #and fit.sigma < prev_sigma:
+            logger.info(f"fix_fingers: best fit found stopping at {idx} {fit.alpha:.2f} << {prev_alpha:0.2f} ")  
             best_fit = fit
             num_fingers = idx - 1
 
             break
         
         if min_alpha is not None and fit.alpha <= min_alpha: 
-            print(f"stopping at min alpha = {fit.alpha:0.2f}")  
+            logger.info(f"stopping at min alpha = {fit.alpha:0.2f}")  
             break
         
         
         prev_fit = fit
         prev_alpha = fit.alpha
         prev_sigma = fit.sigma
-        prev_R = R
+        #prev_R = R
 
             
-    if idx == max_N:
+    if idx == max_fingers:
         logger.info("Unable to find smaller alpha, stopping")
         fit = first_fit
     else:
     #  provide the next 3 alpha, and issue a warning if they are out of whack
-        print("checking alpha")
+        logger.info(f"fix_fingers: checking final alpha, starting with idx={idx}, xmax={xmax}")
         for idy in range(idx+1, idx+4):
-            xmax = np.max(evals[-idy])
-    
+            data = evals[:-idy]
+            if xmax is not None and xmax is not False:
+                xmax = np.max(data)
+                logger.debug(f"fix_fingers: checking final alpha: idy={idx}, xmax={xmax}")
+
+            # TODO: fix , allows xmax settings
+            # if xmax==-1 ?
             if xmin is not None and xmin != -1:
-                check_fit = powerlaw.Fit(evals, xmin=xmin, xmax=xmax, verbose=verbose)
+                check_fit = pl_fit(data=data, xmin=xmin, xmax=xmax, verbose=verbose, distribution=POWER_LAW, pl_package=pl_package)
             else:
-                check_fit = powerlaw.Fit(evals, xmax=xmax, verbose=verbose)
-            logger.info(f"checking fit {idy} xmax {xmax:0.4f}  alpha {fit.alpha:0.2f} sigma {fit.sigma:0.2f}")     
+                check_fit = pl_fit(data=data, xmax=xmax, verbose=verbose, distribution=POWER_LAW, pl_package=pl_package)
+            logger.info(f"checking fit {idy}  alpha {fit.alpha:0.2f} sigma {fit.sigma:0.2f}  xmax {xmax}")     
 
                 
-            
-            if np.abs(check_fit.alpha - fit.alpha) > alpha_thresh/2.0 :
-                logger.warning("clipped fit may be spurious, new alpha found: {check_fit.alpha:0.2f}")
+            if np.abs(check_fit.alpha - fit.alpha) > finger_thresh :
+                logger.debug(f"clipped fit may be spurious, new alpha found: {check_fit.alpha:0.2f}")
                 break
 
-    return best_fit, num_fingers
+    
+    return best_fit, num_fingers, first_fit
              
         
 # https://medium.com/@sourcedexter/how-to-find-the-similarity-between-two-probability-distributions-using-python-a7546e90a08d
@@ -841,3 +992,5 @@ def combine_weights_and_biases(W,b):
             
     return Wb
 
+
+
